import logging
from asyncio import (
    gather,
)
from typing import (
    Any,
)

from aiohttp import (
    ClientConnectorError,
    ClientSession,
)
from yarl import (
    URL,
)

from minos.api_gateway.common import (
    MinosConfig,
)

from ..database import (
    MinosRedisClient,
)
from ..domain.microservice import (
    MICROSERVICE_KEY_PREFIX,
)

logger = logging.getLogger(__name__)


class HealthStatusChecker:
    """Health Status Checker class."""

    def __init__(self, config: MinosConfig):
        self.redis = MinosRedisClient(config=config)

    async def check(self) -> None:
        """Check the health status of the already known microservices.

        :return: This method does not return anything.
        """
        coroutines = []

        async for key in self.redis.redis.scan_iter(match=f"{MICROSERVICE_KEY_PREFIX}:*"):
            coroutines.append(self._check_one(key.decode("utf-8")))

        coroutines = tuple(coroutines)
        await gather(*coroutines)

    async def _check_one(self, key: str):
        logger.info(f"Checking {key!r} health status...")
        try:
            # noinspection PyTypeChecker
            data: dict[str, Any] = await self.redis.get_data(key)
            alive = await self._query_health_status(**data)
            await self._update_one(alive, key, data)
        except Exception as exc:
            logger.warning(f"An exception was raised while checking {key!r}: {exc!r}")

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    async def _query_health_status(self, address: str, port: int, **kwargs) -> bool:
        url = URL.build(scheme="http", host=address, port=port, path="/system/health")

        try:
            async with ClientSession() as session:
                async with session.get(url=url) as response:
                    return response.ok
        except ClientConnectorError:
            return False

    async def _update_one(self, alive: bool, key: str, data: dict[str, Any]) -> None:
        if alive == data["status"]:
            return

        data["status"] = alive
        await self.redis.set_data(key, data)



from __future__ import (
    annotations,
)

from aiomisc.service.periodic import (
    PeriodicService,
)

from minos.api_gateway.common import (
    MinosConfig,
)

from .checkers import (
    HealthStatusChecker,
)


class HealthStatusCheckerService(PeriodicService):
    """Minos DiscoveryPeriodicHealthChecker class."""

    def __init__(self, config: MinosConfig = None, **kwargs):
        super().__init__(**kwargs)
        self.status_checker = HealthStatusChecker(config=config)

    async def callback(self) -> None:
        """Method to be called periodically by the internal ``aiomisc`` logic.

        :return:This method does not return anything.
        """
        await self.status_checker.check()  # pragma: no cover



from __future__ import (
    annotations,
)

from ..exceptions import (
    NotFoundException,
)
from .endpoint import (
    ConcreteEndpoint,
    GenericEndpoint,
)

MICROSERVICE_KEY_PREFIX = "microservice"
ENDPOINT_KEY_PREFIX = "endpoint"


class Microservice:
    """Microservice class."""

    def __init__(self, name: str, address: str, port: int, endpoints: list[list[str]], status: bool = True):
        self.name = name
        self.address = address
        self.port = port
        self.endpoints: list[GenericEndpoint] = [
            GenericEndpoint(endpoint_verb, endpoint_path) for endpoint_verb, endpoint_path in endpoints
        ]
        self.status = status

    @classmethod
    async def find_by_endpoint(cls, concrete_endpoint: ConcreteEndpoint, db_client) -> Microservice:
        """Find a microservice bt endpoint.

        :param concrete_endpoint: The concrete endpoint that must match.
        :param db_client: The database client.
        :return: A ``Microservice`` instance.
        """
        async for key_bytes in db_client.redis.scan_iter(match=f"{ENDPOINT_KEY_PREFIX}:{concrete_endpoint.verb}:*"):
            endpoint = GenericEndpoint.load_by_key(key_bytes)
            if endpoint.matches(concrete_endpoint):
                return await cls.load_by_endpoint(key_bytes, db_client)

        raise NotFoundException

    @classmethod
    async def load_by_endpoint(cls, endpoint_key: bytes, db_client) -> Microservice:
        """Load an instance from the database.

        :param endpoint_key: The endpoint key.
        :param db_client: The database client.
        :return: A ``Microservice`` instance.
        """
        microservice_key = await db_client.get_data(endpoint_key)
        return await cls.load(microservice_key, db_client)

    @classmethod
    async def load(cls, microservice_key: bytes, db_client) -> Microservice:
        """Load an instance from the database.

        :param microservice_key: The microservice key.
        :param db_client: The database client.
        :return: A ``Microservice`` instance.
        """

        microservice_dict = await db_client.get_data(microservice_key)
        microservice_dict["endpoints"] = [
            endpoint_key.split(":", 2)[1:] for endpoint_key in microservice_dict["endpoints"]
        ]
        return cls(**microservice_dict)

    async def save(self, db_client) -> None:
        """Store the instance into the database.

        :param db_client: The database client.
        :return: This method does not return anything.
        """
        microservice_value = {
            "name": self.name,
            "address": self.address,
            "port": self.port,
            "endpoints": [
                f"{ENDPOINT_KEY_PREFIX}:{endpoint.verb}:{endpoint.path_as_str}" for endpoint in self.endpoints
            ],
            "status": self.status,
        }

        microservice_key = f"{MICROSERVICE_KEY_PREFIX}:{self.name}"
        await db_client.set_data(microservice_key, microservice_value)
        for endpoint_key in microservice_value["endpoints"]:
            await db_client.set_data(endpoint_key, microservice_key)

    @classmethod
    async def delete(cls, microservice_name, redis_client):
        microservice_label = f"{MICROSERVICE_KEY_PREFIX}:{microservice_name}"
        microservice_data = await redis_client.get_data(microservice_label)
        endpoints = microservice_data["endpoints"]
        await redis_client.delete_data(microservice_label)
        for endpoint in endpoints:
            await redis_client.delete_data(endpoint)

    def to_json(self):
        microservice_dict = {
            "name": self.name,
            "address": self.address,
            "port": self.port,
            "endpoints": [[endpoint.verb, endpoint.path_as_str] for endpoint in self.endpoints],
            "status": self.status,
        }

        return microservice_dict



import logging
from json import (
    JSONDecodeError,
)

from aiohttp import (
    web,
)

from ..domain import (
    Microservice,
)
from .router import (
    routes,
)

logger = logging.getLogger(__name__)


@routes.view("/microservices/{name}")
class MicroserviceView(web.View):
    async def post(self):
        body = await self.get_body()

        redis_client = self.request.app["db_client"]

        try:
            microservice = Microservice(**body)
        except TypeError as exc:
            raise web.HTTPBadRequest(text=exc.args[0])
        await microservice.save(redis_client)

        return web.HTTPCreated()

    async def delete(self):
        microservice_name = self.request.match_info["name"]

        redis_client = self.request.app["db_client"]

        await Microservice.delete(microservice_name, redis_client)

        return web.HTTPOk()

    async def get_body(self):
        try:
            body = await self.request.json()
        except JSONDecodeError:
            raise web.HTTPBadRequest(text="Empty or wrong body")
        else:
            body["name"] = self.request.match_info["name"]

        return body



import logging
from asyncio import (
    AbstractEventLoop,
)
from typing import (
    NoReturn,
)

from aiomisc.entrypoint import (
    Entrypoint,
)
from aiomisc.utils import (
    create_default_event_loop,
)
from cached_property import (
    cached_property,
)

logger = logging.getLogger(__name__)


class EntrypointLauncher:
    """EntryPoint Launcher class."""

    def __init__(self, services: tuple):
        self.services = services

    def launch(self) -> NoReturn:
        """Launch a new execution and keeps running forever..

        :return: This method does not return anything.
        """
        logger.info("Starting Discovery...")
        try:
            self.loop.run_until_complete(self.entrypoint.__aenter__())
            logger.info("Discovery is up and running!")
            self.loop.run_forever()
        except KeyboardInterrupt:  # pragma: no cover
            logger.info("Stopping discovery...")
        finally:
            self.graceful_shutdown()

    def graceful_shutdown(self, err: Exception = None) -> None:
        """Shutdown the services execution gracefully.

        :return: This method does not return anything.
        """
        self.loop.run_until_complete(self.entrypoint.graceful_shutdown(err))

    @cached_property
    def entrypoint(self) -> Entrypoint:
        """Entrypoint instance.

        :return: An ``Entrypoint`` instance.
        """

        return Entrypoint(*self.services, loop=self.loop)  # pragma: no cover

    @cached_property
    def loop(self) -> AbstractEventLoop:
        """Create the loop.

        :return: An ``AbstractEventLoop`` instance.
        """
        return create_default_event_loop()[0]  # pragma: no cover



__version__ = "0.0.5"

from .cli import (
    app,
)
from .database import (
    MinosRedisClient,
)
from .exceptions import (
    NotFoundException,
)
from .health_status import (
    HealthStatusChecker,
    HealthStatusCheckerService,
)
from .launchers import (
    EntrypointLauncher,
)
from .service import (
    DiscoveryService,
)
